# distutils: include_dirs = dimod/include/

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

__all__ = ['Vartype', 'vartype_info']


cdef extern from "dimod/vartypes.h" namespace "dimod" nogil:
    enum Vartype:
        BINARY
        SPIN
        INTEGER
        REAL

    cdef cppclass vartype_info [Bias]:
        @staticmethod
        Bias default_max(Vartype)
        @staticmethod
        Bias default_min(Vartype)
        @staticmethod
        Bias max(Vartype)
        @staticmethod
        Bias min(Vartype)
